#!/usr/bin/env python 
# -*- coding: utf-8 -*-
"""Runtime NEP calculator wrapper handling CPU/GPU backends."""
import contextlib
import os
import traceback
from collections.abc import Iterable
from pathlib import Path
from typing import Literal

import numpy as np
import numpy.typing as npt
from ase import Atoms
from ase.stress import full_3x3_to_voigt_6_stress
from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.singlepoint import SinglePointCalculator
from loguru import logger
from NepTrainKit.utils import timeit
from NepTrainKit.core import MessageManager
from NepTrainKit.core.structure import Structure
from NepTrainKit.paths import PathLike, as_path
from NepTrainKit.core.types import NepBackend
from NepTrainKit.core.utils import split_by_natoms, aggregate_per_atom_to_structure, is_charge_model
from NepTrainKit.core.cstdio_redirect import redirect_c_stdout_stderr

try:
    from NepTrainKit.nep_cpu import CpuNep
except ImportError:
    logger.debug("no found NepTrainKit.nep_cpu")
    try:
        from nep_cpu import CpuNep
    except ImportError:
        logger.debug("no found nep_cpu")
        CpuNep = None

try:
    from NepTrainKit.nep_gpu import GpuNep
except ImportError:
    logger.debug("no found NepTrainKit.nep_gpu")
    try:
        from nep_gpu import GpuNep
    except ImportError:
        logger.debug("no found nep_gpu")
        GpuNep = None
if GpuNep is not None:
    if "CUDA_VISIBLE_DEVICES" not in os.environ:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"


class NepCalculator:
    """Initialise the NEP calculator and load a CPU/GPU backend."""
    def __init__(
        self,
        model_file: PathLike = "nep.txt",
        backend: NepBackend | None = None,
        batch_size: int | None = None,
        native_stdio: str | Path | Literal["inherit", "silent"] | None = "silent",
    ) -> None:

        super().__init__()
        self.model_path = as_path(model_file)
        if isinstance(backend, str):
            backend = NepBackend(backend)
        self.backend = backend or NepBackend.AUTO
        self.batch_size = batch_size or 1000
        self.initialized = False
        self.nep3 = None
        self.is_charge_model = is_charge_model(self.model_path)
        self.element_list: list[str] = []
        self.type_dict: dict[str, int] = {}
        # Native stdio behavior for C/C++ (printf) in backends
        self._native_stdio = native_stdio
        if CpuNep is None and GpuNep is None:
            MessageManager.send_message_box(
                "Failed to import NEP.\n To use the display functionality normally, please prepare the *.out and descriptor.out files.",
                "Error",
            )
            return
        if self.model_path.exists():
            self.load_nep()
            if getattr(self, "nep3", None) is not None:
                self.element_list = self.nep3.get_element_list()
                self.type_dict = {element: index for index, element in enumerate(self.element_list)}
                self.initialized = True
        else:
            logger.warning(f"NEP model file not found: { self.model_path}" )

    def __del__(self):
        return

    def _native_stdio_ctx(self):
        if self._native_stdio == "inherit":
            return contextlib.nullcontext()
        if self._native_stdio in (None, "silent"):
            return redirect_c_stdout_stderr()
        return redirect_c_stdout_stderr(as_path(self._native_stdio))

    def cancel(self) -> None:
        if getattr(self, "nep3", None) is not None:
            self.nep3.cancel()

    def load_nep(self) -> None:
        if self.backend == NepBackend.AUTO:
            if not self._load_nep_backend(NepBackend.GPU):
                self._load_nep_backend(NepBackend.CPU)
        elif self.backend == NepBackend.GPU:
            if not self._load_nep_backend(NepBackend.GPU):
                MessageManager.send_warning_message("The NEP backend you selected is GPU, but it failed to load on your device; the program has switched to the CPU backend.")
                self._load_nep_backend(NepBackend.CPU)
        else:
            self._load_nep_backend(NepBackend.CPU)

    def _load_nep_backend(self, backend: NepBackend) -> bool:
        try:
            if backend == NepBackend.GPU:
                target_cls = GpuNep
                if target_cls is None:
                    return False
                try:
                    with self._native_stdio_ctx():
                        self.nep3 = target_cls(str(self.model_path))
                        if hasattr(self.nep3, "set_batch_size"):
                            self.nep3.set_batch_size(self.batch_size)
                except RuntimeError as exc:
                    logger.error(exc)
                    MessageManager.send_warning_message(str(exc))
                    return False
            else:

                target_cls=CpuNep
                if target_cls is None:
                    return False
                with self._native_stdio_ctx():
                    self.nep3 = target_cls(str(self.model_path))
            self.backend = backend
            return True
        except Exception:
            logger.debug(traceback.format_exc())
            return False

    @staticmethod
    def _ensure_structure_list(
        structures: Iterable[Structure] | Structure,
    ) -> list[Structure]:
        if isinstance(structures, (Structure, Atoms)):
            return [structures]
        if isinstance(structures, list):
            return structures
        return list(structures)

    @timeit
    def compose_structures(
        self,
        structures: Iterable[Structure] | Structure,
    ) -> tuple[list[list[int]], list[list[float]], list[list[float]], list[int]]:
        structure_list = self._ensure_structure_list(structures)
        group_sizes: list[int] = []
        atom_types: list[list[int]] = []
        boxes: list[list[float]] = []
        positions: list[list[float]] = []
        for structure in structure_list:
            symbols = structure.get_chemical_symbols()
            mapped_types = [self.type_dict[symbol] for symbol in symbols]
            box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            coords = structure.positions.transpose(1, 0).reshape(-1).tolist()
            atom_types.append(mapped_types)
            boxes.append(box)
            positions.append(coords)
            group_sizes.append(len(mapped_types))
        return atom_types, boxes, positions, group_sizes

    @timeit
    def calculate(
        self,
        structures: Iterable[Structure] | Structure,
        return_charge: bool = False,
        mean_virial: bool = True,
    ):
        structure_list = self._ensure_structure_list(structures)
        if not self.initialized or not structure_list:
            empty = np.array([], dtype=np.float32)
            if self.is_charge_model and return_charge:
                return empty, [], [], [], []
            return empty, [], []
        atom_types, boxes, positions, group_sizes = self.compose_structures(structure_list)
        self.nep3.reset_cancel()
        try:
            with self._native_stdio_ctx():
                if self.is_charge_model:
                    if not hasattr(self.nep3, "calculate_qnep"):
                        raise RuntimeError("Charge model backend does not implement calculate_qnep().")
                    outputs = self.nep3.calculate_qnep(atom_types, boxes, positions)
                else:
                    outputs = self.nep3.calculate(atom_types, boxes, positions)
        except Exception as exc:
            logger.error(exc)
            MessageManager.send_warning_message(str(exc))
            empty = np.array([], dtype=np.float32)
            if self.is_charge_model and return_charge:
                return empty, [], [], [], []
            return empty, [], []

        if self.is_charge_model:
            potentials, forces, virials, charges, becs = outputs
            pot_arr = np.asarray(potentials)
            if pot_arr.dtype != object:
                pot_arr = pot_arr.astype(np.float32, copy=False)
                forces_arr = np.asarray(forces, dtype=np.float32)
                virials_arr = np.asarray(virials, dtype=np.float32)
                charges_arr = np.asarray(charges, dtype=np.float32)
                becs_arr = np.asarray(becs, dtype=np.float32)
                if forces_arr.ndim == 1:
                    forces_arr = forces_arr.reshape(-1, 3)
                if virials_arr.ndim == 1:
                    virials_arr = virials_arr.reshape(-1, 9)
                if charges_arr.ndim > 1:
                    charges_arr = charges_arr.reshape(-1)
                if becs_arr.ndim == 1:
                    becs_arr = becs_arr.reshape(-1, 9)
                potentials_array = aggregate_per_atom_to_structure(pot_arr, group_sizes, map_func=np.sum, axis=None).tolist()
                force_blocks = split_by_natoms(forces_arr, group_sizes)
                virial_blocks = aggregate_per_atom_to_structure(virials_arr, group_sizes, map_func=np.mean, axis=0).tolist()
                charge_blocks = [blk.reshape(-1) for blk in split_by_natoms(charges_arr, group_sizes)]
                bec_blocks = split_by_natoms(becs_arr.reshape(-1, 9), group_sizes)
                if not return_charge:
                    return potentials_array, force_blocks, virial_blocks
                return potentials_array, force_blocks, virial_blocks, charge_blocks, bec_blocks
            else:
                potentials = np.hstack(potentials) if len(potentials) else np.array([])
                potentials_array = aggregate_per_atom_to_structure(potentials, group_sizes, map_func=np.sum, axis=None)
                reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]
                reshaped_virials = [np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials]
                if not return_charge:
                    return potentials_array.tolist(), reshaped_forces, reshaped_virials

                charges_list = [np.asarray(charge, dtype=np.float32).reshape(-1) for charge in (charges or [])]
                becs_list = [np.asarray(bec, dtype=np.float32).reshape(9, -1).T for bec in (becs or [])]
                return potentials_array.tolist(), reshaped_forces, reshaped_virials, charges_list, becs_list

        potentials, forces, virials = outputs
        potentials_arr = np.asarray(potentials, dtype=np.float32)
        forces_arr = np.asarray(forces, dtype=np.float32)
        virials_arr = np.asarray(virials, dtype=np.float32)
        if potentials_arr.size == 0:
            return [], [], []
        if forces_arr.ndim == 1:
            forces_arr = forces_arr.reshape(-1, 3)
        if virials_arr.ndim == 1:
            virials_arr = virials_arr.reshape(-1, 9)
        potentials_array = aggregate_per_atom_to_structure(potentials_arr, group_sizes, map_func=np.sum, axis=None).tolist()
        forces_blocks = split_by_natoms(forces_arr, group_sizes)
        if mean_virial:
            virials_blocks = aggregate_per_atom_to_structure(virials_arr, group_sizes, map_func=np.mean, axis=0).tolist()
        else:
            virials_blocks = split_by_natoms(virials_arr, group_sizes)

        return potentials_array, forces_blocks, virials_blocks

    @timeit
    def calculate_dftd3(
        self,
        structures: Iterable[Structure] | Structure,
        functional: str,
        cutoff: float,
        cutoff_cn: float,
        mean_virial: bool = True,
    ):
        structure_list = self._ensure_structure_list(structures)
        if not self.initialized or not structure_list:
            return [], [], []
        atom_types, boxes, positions, group_sizes = self.compose_structures(structure_list)
        self.nep3.reset_cancel()
        with self._native_stdio_ctx():
            potentials, forces, virials = self.nep3.calculate_dftd3(
                functional,
                cutoff,
                cutoff_cn,
                atom_types,
                boxes,
                positions,
            )
        potentials_arr = np.asarray(potentials, dtype=np.float32)
        forces_arr = np.asarray(forces, dtype=np.float32)
        virials_arr = np.asarray(virials, dtype=np.float32)
        if forces_arr.ndim == 1:
            forces_arr = forces_arr.reshape(-1, 3)
        if virials_arr.ndim == 1:
            virials_arr = virials_arr.reshape(-1, 9)
        potentials_array = aggregate_per_atom_to_structure(potentials_arr, group_sizes, map_func=np.sum, axis=None).tolist()
        forces_blocks = split_by_natoms(forces_arr, group_sizes)
        if mean_virial:
            virials_blocks = aggregate_per_atom_to_structure(virials_arr, group_sizes, map_func=np.mean, axis=0).tolist()
        else:
            virials_blocks = split_by_natoms(virials_arr, group_sizes)
        return potentials_array, forces_blocks, virials_blocks

    @timeit
    def calculate_with_dftd3(
        self,
        structures: Iterable[Structure] | Structure,
        functional: str,
        cutoff: float,
        cutoff_cn: float,
        mean_virial: bool = True,
    ):
        structure_list = self._ensure_structure_list(structures)
        if not self.initialized or not structure_list:
            return [], [], []
        atom_types, boxes, positions, group_sizes = self.compose_structures(structure_list)
        self.nep3.reset_cancel()
        with self._native_stdio_ctx():
            potentials, forces, virials = self.nep3.calculate_with_dftd3(
                functional,
                cutoff,
                cutoff_cn,
                atom_types,
                boxes,
                positions,
            )

        potentials_arr = np.asarray(potentials, dtype=np.float32)
        forces_arr = np.asarray(forces, dtype=np.float32)
        virials_arr = np.asarray(virials, dtype=np.float32)
        if forces_arr.ndim == 1:
            forces_arr = forces_arr.reshape(-1, 3)
        if virials_arr.ndim == 1:
            virials_arr = virials_arr.reshape(-1, 9)
        potentials_array = aggregate_per_atom_to_structure(potentials_arr, group_sizes, map_func=np.sum, axis=None).tolist()
        forces_blocks = split_by_natoms(forces_arr, group_sizes)
        if mean_virial:
            virials_blocks = aggregate_per_atom_to_structure(virials_arr, group_sizes, map_func=np.mean, axis=0).tolist()
        else:
            virials_blocks = split_by_natoms(virials_arr, group_sizes)
        return potentials_array, forces_blocks, virials_blocks

    def get_descriptor(self, structure: Structure) -> npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        return self.get_structures_descriptor([structure], mean_descriptor=False)

    @timeit
    def get_structures_descriptor(
        self,
        structures: list[Structure],
        mean_descriptor: bool = True
    ) -> npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        types, boxes, positions, group_sizes = self.compose_structures(structures)
        self.nep3.reset_cancel()
        with self._native_stdio_ctx():
            descriptor = self.nep3.get_structures_descriptor(types, boxes, positions)
        descriptor = np.asarray(descriptor, dtype=np.float32)
        if descriptor.size == 0:
            return descriptor
        if not mean_descriptor:
            return descriptor
        structure_descriptor = aggregate_per_atom_to_structure(descriptor, group_sizes, map_func=np.mean, axis=0)
        return structure_descriptor

    @timeit
    def get_structures_polarizability(
        self,
        structures: list[Structure],
    ) -> npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        types, boxes, positions, _ = self.compose_structures(structures)
        self.nep3.reset_cancel()

        with self._native_stdio_ctx():
            polarizability = self.nep3.get_structures_polarizability(types, boxes, positions)
        return np.array(polarizability, dtype=np.float32)

    def get_structures_dipole(
        self,
        structures: list[Structure],
    ) -> npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        self.nep3.reset_cancel()

        types, boxes, positions, _ = self.compose_structures(structures)
        with self._native_stdio_ctx():
            dipole = self.nep3.get_structures_dipole(types, boxes, positions)
        return np.array(dipole, dtype=np.float32)

    def calculate_to_ase(
            self,
            atoms_list: Atoms | Iterable[Atoms],
            calc_descriptor=False,

    ):
        if isinstance(atoms_list, Atoms):
            atoms_list = [atoms_list]
        descriptor_blocks: list[np.ndarray] | None = None
        if calc_descriptor:
            per_atom_descriptor = self.get_structures_descriptor(atoms_list, mean_descriptor=False)
            atom_counts = [len(atoms) for atoms in atoms_list]
            descriptor_blocks = split_by_natoms(per_atom_descriptor, atom_counts)

        energy, forces, virial = self.calculate(atoms_list)

        for index, atoms in enumerate(atoms_list):
            _e = energy[index]
            _f = forces[index]
            _vi = np.asarray(virial[index])
            if _vi.ndim == 2 and _vi.shape[1] == 9:
                _vi_avg = _vi.mean(axis=0)
            else:
                _vi_avg = _vi.reshape(9)
            _s = _vi_avg.reshape(3, 3) * len(atoms) / atoms.get_volume()
            spc = SinglePointCalculator(
                atoms,
                energy=_e,
                forces=_f,
                stress=full_3x3_to_voigt_6_stress(_s),

            )
            if calc_descriptor:
                spc.results["descriptor"] = descriptor_blocks[index]
            atoms.calc = spc


Nep3Calculator = NepCalculator


class NepAseCalculator(Calculator):
    implemented_properties=[
        "energy",
        "energies",
        "forces",
        "stress",
        "descriptor",
    ]
    def __init__(self,
                 model_file: PathLike = "nep.txt",
                backend: NepBackend | None = None,
                batch_size: int | None = None,*args,**kwargs) -> None:

        self._calc=NepCalculator(model_file,backend,batch_size)
        Calculator.__init__(self,*args,**kwargs)

    def calculate(
        self, atoms=None, properties=['energy'], system_changes=all_changes
    ):

        if properties is None:
            properties = self.implemented_properties
        super().calculate(atoms,properties,system_changes)
        if "descriptor" in properties:
            descriptor = self._calc.get_descriptor(atoms)
            self.results["descriptor"]=descriptor
        energy,forces,virial = self._calc.calculate(atoms)

        self.results["energy"]=energy[0]
        self.results["forces"]=forces[0]
        virial=np.asarray(virial[0])
        if virial.ndim == 2 and virial.shape[1] == 9:
            virial_val = virial.mean(axis=0)
        else:
            virial_val = virial.reshape(9)
        stress = virial_val.reshape(3,3)*len(atoms)/atoms.get_volume()
        self.results["stress"]=full_3x3_to_voigt_6_stress(stress)



"""Input widgets that pair spin boxes with unit labels."""

from __future__ import annotations

from PySide6.QtWidgets import QFrame, QHBoxLayout, QSpinBox, QDoubleSpinBox
from qfluentwidgets import BodyLabel


class SpinBoxUnitInputFrame(QFrame):
    """Composite input frame with spin boxes followed by unit labels."""

    def __init__(self, parent=None):
        """Create the layout and track added input widgets.

        Parameters
        ----------
        parent : QWidget, optional
            Parent widget responsible for ownership.
        """
        super(SpinBoxUnitInputFrame, self).__init__(parent)
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.object_list: list[QSpinBox | QDoubleSpinBox] = []

    def set_input(self, unit_str, object_num, input_type="int"):
        """Populate the frame with spin boxes and unit labels.

        Parameters
        ----------
        unit_str : str or list[str]
            Unit string applied to each input or list of per-input units.
        object_num : int
            Number of inputs to create.
        input_type : str or list[str], optional
            Either a single type of "int"/"float" or a list specifying the
            type for each input in sequence.

        Raises
        ------
        TypeError
            Raised when `unit_str` or `input_type` is not a string or list.
        """
        if isinstance(unit_str, str):
            unit_str = [unit_str] * object_num
        elif isinstance(unit_str, list):
            unit_str = unit_str
        else:
            raise TypeError("unit_str must be str or list")

        if isinstance(input_type, str):
            input_type = [input_type] * object_num
        elif isinstance(input_type, list):
            input_type = input_type
        else:
            raise TypeError("input_type must be str or list")

        for i in range(object_num):
            if input_type[i % len(unit_str)] == "int":
                input_object = QSpinBox(self)
                input_object.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)
            elif input_type[i % len(unit_str)] == "float":
                input_object = QDoubleSpinBox(self)
                input_object.setDecimals(3)
                input_object.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)
            else:
                raise TypeError("input_type must be int or float")

            input_object.setFixedHeight(25)
            self._layout.addWidget(input_object)
            self._layout.addWidget(BodyLabel(unit_str[i % len(unit_str)], self))
            self.object_list.append(input_object)

    def setRange(self, min_value, max_value):
        """Apply the same range constraints to every spin box.

        Parameters
        ----------
        min_value : int | float
            Minimum allowable value for the inputs.
        max_value : int | float
            Maximum allowable value for the inputs.
        """
        for input_object in self.object_list:
            input_object.setRange(min_value, max_value)

    def setDecimals(self, decimals: int):
        """Set the number of decimals for every double spin box.

        Parameters
        ----------
        decimals : int
            Number of decimal places.
        """
        for input_object in self.object_list:
            if isinstance(input_object, QDoubleSpinBox):
                input_object.setDecimals(decimals)

    def get_input_value(self) -> list[int | float]:
        """Return the numeric values from each input widget.

        Returns
        -------
        list[int | float]
            Values retrieved from the spin boxes in order.
        """
        return [input_object.value() for input_object in self.object_list]

    def set_input_value(self, value_list):
        """Populate the spin boxes with supplied values.

        Parameters
        ----------
        value_list : int or float or list[int | float]
            Single value or list of values applied to the inputs.
        """
        if not isinstance(value_list, list):
            value_list = [value_list] * len(self.object_list)

        for i, input_object in enumerate(self.object_list):
            input_object.setValue(value_list[i])




"""Card that wraps the torsion-guard PBC configurator for organic molecules."""

from __future__ import annotations

from typing import Any, List

import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox

from NepTrainKit.core import CardManager
from NepTrainKit.core.torsion_guard_pbc import (
    TorsionGuardParams,
    process_single as tg_process_single,
)
from NepTrainKit.ui.widgets import SpinBoxUnitInputFrame
from NepTrainKit.ui.widgets import MakeDataCard


@CardManager.register_card
class OrganicMolConfigPBCCard(MakeDataCard):
    """Create torsion-driven molecular configurations using the TorsionGuard PBC workflow.
    
    Parameters
    ----------
    parent : QWidget, optional
        Parent widget that owns the configuration card.
    """

    group = "Perturbation"
    card_name = "Organic Mol Config"
    menu_icon = r":/images/src/images/perturb.svg"

    def __init__(self, parent=None):
        """Initialise the card and build its configuration widgets.
        
        Parameters
        ----------
        parent : QWidget, optional
            Parent widget passed to the base card constructor.
        """
        super().__init__(parent)
        self.setTitle("Organic Molecular Configuration(Zherui Chen)")
        self._init_ui()

    # ---------- UI ----------
    def _init_ui(self):
        """Create all of the widgets required to configure the torsion-guard workflow.
        """
        self.setObjectName("organic_mol_config_pbc_card")

        row = 0

        # perturb_per_frame
        self.perturb_label = BodyLabel("Confs per structure:", self.setting_widget)
        self.perturb_label.setToolTip("Number of perturbed conformations generated per input structure")
        self.perturb_label.installEventFilter(ToolTipFilter(self.perturb_label, 300, ToolTipPosition.TOP))
        self.perturb_frame = SpinBoxUnitInputFrame(self)
        self.perturb_frame.set_input("count", 1, "int")
        self.perturb_frame.setRange(1, 100000)
        self.perturb_frame.set_input_value([100])
        self.settingLayout.addWidget(self.perturb_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.perturb_frame, row, 1, 1, 2)
        row += 1

        # torsion_range_deg
        self.torsion_label = BodyLabel("Torsion range:", self.setting_widget)
        self.torsion_label.setToolTip("Torsion angle range (degrees)")
        self.torsion_label.installEventFilter(ToolTipFilter(self.torsion_label, 300, ToolTipPosition.TOP))
        self.torsion_frame = SpinBoxUnitInputFrame(self)
        self.torsion_frame.set_input(["°", "°"], 2, ["float", "float"])
        self.torsion_frame.setRange(-360, 360)
        self.torsion_frame.set_input_value([-180.0, 180.0])
        self.settingLayout.addWidget(self.torsion_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.torsion_frame, row, 1, 1, 2)
        row += 1

        # max_torsions_per_conf
        self.max_torsions_label = BodyLabel("Max torsions/conf:", self.setting_widget)
        self.max_torsions_label.setToolTip("Maximum number of torsions applied per conformation")
        self.max_torsions_label.installEventFilter(ToolTipFilter(self.max_torsions_label, 300, ToolTipPosition.TOP))
        self.max_torsions_frame = SpinBoxUnitInputFrame(self)
        self.max_torsions_frame.set_input("", 1, "int")
        self.max_torsions_frame.setRange(0, 10000)
        self.max_torsions_frame.set_input_value([50])
        self.settingLayout.addWidget(self.max_torsions_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.max_torsions_frame, row, 1, 1, 2)
        row += 1

        # gaussian_sigma
        self.sigma_label = BodyLabel("Gaussian sigma:", self.setting_widget)
        self.sigma_label.setToolTip("Std dev of added Gaussian noise (Å)")
        self.sigma_label.installEventFilter(ToolTipFilter(self.sigma_label, 300, ToolTipPosition.TOP))
        self.sigma_frame = SpinBoxUnitInputFrame(self)
        self.sigma_frame.set_input("Å", 1, "float")
        self.sigma_frame.setRange(0, 5)
        self.sigma_frame.set_input_value([0.03])
        self.settingLayout.addWidget(self.sigma_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.sigma_frame, row, 1, 1, 2)
        row += 1

        # pbc mode
        self.pbc_label = BodyLabel("PBC mode:", self.setting_widget)
        self.pbc_label.setToolTip("auto: use cell if present; yes: force PBC; no: non-PBC")
        self.pbc_label.installEventFilter(ToolTipFilter(self.pbc_label, 300, ToolTipPosition.TOP))
        self.pbc_combo = ComboBox(self.setting_widget)
        for opt in ("auto", "yes", "no"):
            self.pbc_combo.addItem(opt)
        self.pbc_combo.setCurrentIndex(0)
        self.settingLayout.addWidget(self.pbc_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.pbc_combo, row, 1, 1, 2)
        row += 1

        # local_mode_cutoff_atoms
        self.local_cut_label = BodyLabel("Local-mode cutoff atoms:", self.setting_widget)
        self.local_cut_label.setToolTip("Use local subtree rotations if atoms > this threshold")
        self.local_cut_label.installEventFilter(ToolTipFilter(self.local_cut_label, 300, ToolTipPosition.TOP))
        self.local_cut_frame = SpinBoxUnitInputFrame(self)
        self.local_cut_frame.set_input("atoms", 1, "int")
        self.local_cut_frame.setRange(0, 1000000)
        self.local_cut_frame.set_input_value([200])
        self.settingLayout.addWidget(self.local_cut_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.local_cut_frame, row, 1, 1, 2)
        row += 1

        # local_torsion_max_subtree
        self.local_sub_label = BodyLabel("Max subtree size:", self.setting_widget)
        self.local_sub_label.setToolTip("Maximum atoms in rotated subtree for local mode")
        self.local_sub_label.installEventFilter(ToolTipFilter(self.local_sub_label, 300, ToolTipPosition.TOP))
        self.local_sub_frame = SpinBoxUnitInputFrame(self)
        self.local_sub_frame.set_input("atoms", 1, "int")
        self.local_sub_frame.setRange(1, 100000)
        self.local_sub_frame.set_input_value([100])
        self.settingLayout.addWidget(self.local_sub_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.local_sub_frame, row, 1, 1, 2)
        row += 1

        # bond_detect_factor
        self.bond_detect_label = BodyLabel("Bond detect factor:", self.setting_widget)
        self.bond_detect_label.setToolTip("Bond detection cutoff multiplier (ri+rj)")
        self.bond_detect_label.installEventFilter(ToolTipFilter(self.bond_detect_label, 300, ToolTipPosition.TOP))
        self.bond_detect_frame = SpinBoxUnitInputFrame(self)
        self.bond_detect_frame.set_input("x", 1, "float")
        self.bond_detect_frame.setRange(0, 5)
        self.bond_detect_frame.set_input_value([1.15])
        self.settingLayout.addWidget(self.bond_detect_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.bond_detect_frame, row, 1, 1, 2)
        row += 1

        # bond_keep_min_factor
        self.bond_min_label = BodyLabel("Bond min factor:", self.setting_widget)
        self.bond_min_label.setToolTip("Lower bound for bonded distances; 0 disables")
        self.bond_min_label.installEventFilter(ToolTipFilter(self.bond_min_label, 300, ToolTipPosition.TOP))
        self.bond_min_frame = SpinBoxUnitInputFrame(self)
        self.bond_min_frame.set_input("x", 1, "float")
        self.bond_min_frame.setRange(0, 5)
        self.bond_min_frame.set_input_value([0.60])
        self.settingLayout.addWidget(self.bond_min_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.bond_min_frame, row, 1, 1, 2)
        row += 1

        # Pauling bond-order params
        self.bo_c_label = BodyLabel("Pauling c constant:", self.setting_widget)
        self.bo_c_label.setToolTip("Bond order constant c in exp((r0-r)/c)")
        self.bo_c_label.installEventFilter(ToolTipFilter(self.bo_c_label, 300, ToolTipPosition.TOP))
        self.bo_c_frame = SpinBoxUnitInputFrame(self)
        self.bo_c_frame.set_input("", 1, "float")
        self.bo_c_frame.setRange(0.01, 2.0)
        self.bo_c_frame.set_input_value([0.3])
        self.settingLayout.addWidget(self.bo_c_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.bo_c_frame, row, 1, 1, 2)
        row += 1

        self.bo_thr_label = BodyLabel("BondOrder threshold:", self.setting_widget)
        self.bo_thr_label.setToolTip("Minimum bond order to form bond (default 0.2)")
        self.bo_thr_label.installEventFilter(ToolTipFilter(self.bo_thr_label, 300, ToolTipPosition.TOP))
        self.bo_thr_frame = SpinBoxUnitInputFrame(self)
        self.bo_thr_frame.set_input("", 1, "float")
        self.bo_thr_frame.setRange(0.0, 1.0)
        self.bo_thr_frame.set_input_value([0.2])
        self.settingLayout.addWidget(self.bo_thr_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.bo_thr_frame, row, 1, 1, 2)
        row += 1

        # bond_keep_max_factor (optional)
        self.bond_max_label = BodyLabel("Bond max factor:", self.setting_widget)
        self.bond_max_label.setToolTip("Upper bound for bonded distances; uncheck to disable")
        self.bond_max_label.installEventFilter(ToolTipFilter(self.bond_max_label, 300, ToolTipPosition.TOP))
        self.bond_max_frame = SpinBoxUnitInputFrame(self)
        self.bond_max_frame.set_input("x", 1, "float")
        self.bond_max_frame.setRange(0, 5)
        self.bond_max_frame.set_input_value([1.15])
        self.bond_max_enable = CheckBox("Enable upper bound", self.setting_widget)
        self.bond_max_enable.setChecked(False)
        self.settingLayout.addWidget(self.bond_max_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.bond_max_frame, row, 1, 1, 1)
        self.settingLayout.addWidget(self.bond_max_enable, row, 2, 1, 1)
        row += 1

        # nonbond_min_factor
        self.nonbond_min_label = BodyLabel("Non-bonded min factor:", self.setting_widget)
        self.nonbond_min_label.setToolTip("Minimum separation for non-bonded atoms (ri+rj) factor")
        self.nonbond_min_label.installEventFilter(ToolTipFilter(self.nonbond_min_label, 300, ToolTipPosition.TOP))
        self.nonbond_min_frame = SpinBoxUnitInputFrame(self)
        self.nonbond_min_frame.set_input("x", 1, "float")
        self.nonbond_min_frame.setRange(0, 5)
        self.nonbond_min_frame.set_input_value([0.80])
        self.settingLayout.addWidget(self.nonbond_min_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.nonbond_min_frame, row, 1, 1, 2)
        row += 1

        # max_retries_per_frame
        self.retries_label = BodyLabel("Max retries:", self.setting_widget)
        self.retries_label.setToolTip("Backoff retries per conformation if guards fail")
        self.retries_label.installEventFilter(ToolTipFilter(self.retries_label, 300, ToolTipPosition.TOP))
        self.retries_frame = SpinBoxUnitInputFrame(self)
        self.retries_frame.set_input("tries", 1, "int")
        self.retries_frame.setRange(0, 100)
        self.retries_frame.set_input_value([12])
        self.settingLayout.addWidget(self.retries_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.retries_frame, row, 1, 1, 2)
        row += 1

        # MULT_BOND_FACTOR
        self.multbond_label = BodyLabel("Multi-bond factor:", self.setting_widget)
        self.multbond_label.setToolTip("Exclude suspected multiple bonds if d < factor*(ri+rj)")
        self.multbond_label.installEventFilter(ToolTipFilter(self.multbond_label, 300, ToolTipPosition.TOP))
        self.multbond_frame = SpinBoxUnitInputFrame(self)
        self.multbond_frame.set_input("x", 1, "float")
        self.multbond_frame.setRange(0, 2)
        self.multbond_frame.set_input_value([0.87])
        self.settingLayout.addWidget(self.multbond_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.multbond_frame, row, 1, 1, 2)
        row += 1

        # nonpbc_box_size
        self.box_label = BodyLabel("Non-PBC box size:", self.setting_widget)
        self.box_label.setToolTip("Box edge for non-periodic output (Å)")
        self.box_label.installEventFilter(ToolTipFilter(self.box_label, 300, ToolTipPosition.TOP))
        self.box_frame = SpinBoxUnitInputFrame(self)
        self.box_frame.set_input("Å", 1, "float")
        self.box_frame.setRange(1, 100000)
        self.box_frame.set_input_value([100.0])
        self.settingLayout.addWidget(self.box_label, row, 0, 1, 1)
        self.settingLayout.addWidget(self.box_frame, row, 1, 1, 2)

    def _current_pbc_mode(self) -> str:
        """Return the currently selected periodic boundary mode.
        
        Returns
        -------
        str
            One of ``"auto"``, ``"yes"``, or ``"no"``.
        """
        return self.pbc_combo.currentText()

    # ---------- Core ----------
    def process_structure(self, structure) -> list[Any]:
        """Generate torsion-driven molecular conformers using the TorsionGuard PBC workflow.
        
        Parameters
        ----------
        structure : ase.Atoms
            Structure providing the initial molecular coordinates and cell.
        
        Returns
        -------
        list[ase.Atoms]
            Structures returned by the torsion-guard generator.
        """
        # structure is an ASE Atoms
        from ase import Atoms  # local import to avoid hard dep in module import time

        pbc_mode = self._current_pbc_mode()

        symbols: List[str] = structure.get_chemical_symbols()
        coords: np.ndarray = structure.get_positions().astype(float)

        # Decide cell: always pass a 3x3 even for non-PBC if available; module decides using pbc_mode
        cell_mat = None
        try:
            cell_arr = structure.get_cell().array  # type: ignore[attr-defined]
            if cell_arr is not None and np.array(cell_arr).shape == (3, 3):
                cell_mat = np.array(cell_arr, dtype=float)
        except Exception:
            cell_mat = None

        params = TorsionGuardParams(
            perturb_per_frame=int(self.perturb_frame.get_input_value()[0]),
            torsion_range_deg=tuple(map(float, self.torsion_frame.get_input_value())),
            max_torsions_per_conf=int(self.max_torsions_frame.get_input_value()[0]),
            gaussian_sigma=float(self.sigma_frame.get_input_value()[0]),
            pbc_mode=pbc_mode,
            local_mode_cutoff_atoms=int(self.local_cut_frame.get_input_value()[0]),
            local_torsion_max_subtree=int(self.local_sub_frame.get_input_value()[0]),
            bond_detect_factor=float(self.bond_detect_frame.get_input_value()[0]),
            bond_keep_min_factor=float(self.bond_min_frame.get_input_value()[0]),
            bond_keep_max_factor=float(self.bond_max_frame.get_input_value()[0]) if self.bond_max_enable.isChecked() else None,
            nonbond_min_factor=float(self.nonbond_min_frame.get_input_value()[0]),
            max_retries_per_frame=int(self.retries_frame.get_input_value()[0]),
            mult_bond_factor=float(self.multbond_frame.get_input_value()[0]),
            nonpbc_box_size=float(self.box_frame.get_input_value()[0]),
            bo_c_const=float(self.bo_c_frame.get_input_value()[0]),
            bo_threshold=float(self.bo_thr_frame.get_input_value()[0]),
        )

        result_list = tg_process_single(symbols, coords, cell_mat, params)

        structures_out: list[Atoms] = []
        for sym, new_coords, cell, pbc_active in result_list:
            new_atoms: Atoms = structure.copy()
            new_atoms.set_positions(np.array(new_coords, dtype=float))

            # Set periodicity and cell
            if pbc_active and cell is not None:
                new_atoms.set_cell(np.array(cell, dtype=float))
                new_atoms.set_pbc(True)
                try:
                    new_atoms.wrap()
                except Exception:
                    pass
            else:
                # Centered in a non-PBC box already by module; attach a large box for visualization
                L = float(self.box_frame.get_input_value()[0])
                new_atoms.set_cell(np.diag([L, L, L]))
                new_atoms.set_pbc(False)

            # Tagging
            cfg = new_atoms.info.get("Config_type", "")
            cfg += f" TorsionGuard(n={self.perturb_frame.get_input_value()[0]}, sigma={self.sigma_frame.get_input_value()[0]}, pbc={pbc_mode})"
            new_atoms.info["Config_type"] = cfg.strip()

            structures_out.append(new_atoms)

        return structures_out

    # ---------- Persistence ----------
    def to_dict(self):
        """Serialize the current configuration to a plain dictionary.
        
        Returns
        -------
        dict
            Dictionary that can be fed into ``from_dict`` to rebuild the state.
        """
        data = super().to_dict()
        data.update({
            "perturb_per_frame": self.perturb_frame.get_input_value(),
            "torsion_range_deg": self.torsion_frame.get_input_value(),
            "max_torsions_per_conf": self.max_torsions_frame.get_input_value(),
            "gaussian_sigma": self.sigma_frame.get_input_value(),
            "pbc_mode": self.pbc_combo.currentText(),
            "local_cutoff": self.local_cut_frame.get_input_value(),
            "local_subtree": self.local_sub_frame.get_input_value(),
            "bond_detect_factor": self.bond_detect_frame.get_input_value(),
            "bond_keep_min_factor": self.bond_min_frame.get_input_value(),
            "bond_keep_max_factor": self.bond_max_frame.get_input_value(),
            "bond_keep_max_enable": self.bond_max_enable.isChecked(),
            "nonbond_min_factor": self.nonbond_min_frame.get_input_value(),
            "max_retries": self.retries_frame.get_input_value(),
            "mult_bond_factor": self.multbond_frame.get_input_value(),
            "nonpbc_box_size": self.box_frame.get_input_value(),
            "bo_c_const": self.bo_c_frame.get_input_value(),
            "bo_threshold": self.bo_thr_frame.get_input_value(),
        })
        return data

    def from_dict(self, data_dict):
        """Restore the card configuration from serialized values.
        
        Parameters
        ----------
        data_dict : dict
            Serialized configuration previously produced by ``to_dict``.
        """
        super().from_dict(data_dict)
        self.perturb_frame.set_input_value(data_dict.get("perturb_per_frame", [100]))
        self.torsion_frame.set_input_value(data_dict.get("torsion_range_deg", [-180.0, 180.0]))
        self.max_torsions_frame.set_input_value(data_dict.get("max_torsions_per_conf", [5]))
        self.sigma_frame.set_input_value(data_dict.get("gaussian_sigma", [0.03]))

        pbc_mode = data_dict.get("pbc_mode", "auto")
        idx = {"auto": 0, "yes": 1, "no": 2}.get(pbc_mode, 0)
        self.pbc_combo.setCurrentIndex(idx)

        self.local_cut_frame.set_input_value(data_dict.get("local_cutoff", [150]))
        self.local_sub_frame.set_input_value(data_dict.get("local_subtree", [40]))
        self.bond_detect_frame.set_input_value(data_dict.get("bond_detect_factor", [1.15]))
        self.bond_min_frame.set_input_value(data_dict.get("bond_keep_min_factor", [0.60]))
        self.bond_max_frame.set_input_value(data_dict.get("bond_keep_max_factor", [1.15]))
        self.bond_max_enable.setChecked(bool(data_dict.get("bond_keep_max_enable", False)))
        self.nonbond_min_frame.set_input_value(data_dict.get("nonbond_min_factor", [0.80]))
        self.retries_frame.set_input_value(data_dict.get("max_retries", [12]))
        self.multbond_frame.set_input_value(data_dict.get("mult_bond_factor", [0.87]))
        self.box_frame.set_input_value(data_dict.get("nonpbc_box_size", [100.0]))
        self.bo_c_frame.set_input_value(data_dict.get("bo_c_const", [0.3]))
        self.bo_thr_frame.set_input_value(data_dict.get("bo_threshold", [0.2]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import os
import platform
from pathlib import Path
from typing import Iterator, Sequence


PathLike = str | Path


def as_path(value: PathLike | None, *, base: Path | None = None) -> Path:
    """Return ``value`` as a :class:`Path`, optionally relative to ``base``.

    - If ``value`` is ``None``, return ``base`` as a Path (or raise if base is None).
    - If ``value`` is relative and ``base`` is provided, join with ``base``.
    """
    if value is None:
        if base is None:
            raise ValueError('value and base cannot both be None')
        return Path(base)
    path_value = Path(value)
    if base and not path_value.is_absolute():
        return Path(base) / path_value
    return path_value


def ensure_directory(path: PathLike, *, exist_ok: bool = True) -> Path:
    """Create ``path`` if needed and return it as a :class:`Path`."""
    directory = Path(path)
    directory.mkdir(parents=True, exist_ok=exist_ok)
    return directory


def ensure_suffix(path: PathLike, suffix: str) -> Path:
    """Return ``path`` with the provided ``suffix`` appended if missing."""
    p = Path(path)
    if not suffix.startswith('.'):
        suffix = f'.{suffix}'
    return p.with_suffix(suffix) if p.suffix != suffix else p


def iter_files(root: PathLike, patterns: Sequence[str] | None = None) -> Iterator[Path]:
    """Yield files under ``root`` optionally filtered by glob ``patterns``."""
    base = Path(root)
    if not base.exists():
        return iter(())
    if not patterns:
        return base.rglob('*')

    def generator() -> Iterator[Path]:
        for pattern in patterns:
            yield from base.rglob(pattern)

    return generator()


def check_path_type(path: str | Path) -> str:
    """Return ``folder`` or ``file`` even if ``path`` does not exist."""
    p = Path(path)
    if p.is_dir():
        return "folder"
    if p.is_file():
        return "file"
    return "file" if p.suffix else "folder"


def get_user_config_path() -> Path:
    """Return the user-specific configuration path as a ``Path`` instance."""
    if platform.system() == 'Windows':
        local = os.getenv('LOCALAPPDATA')
        if not local:
            user_profile = os.getenv('USERPROFILE')
            local_path = as_path(user_profile) / 'AppData' / 'Local' if user_profile else Path.home() / 'AppData' / 'Local'
        else:
            local_path = as_path(local)
        return ensure_directory(local_path / 'NepTrainKit')

    return ensure_directory(Path.home() / '.config' / 'NepTrainKit')


__all__ = [
    'PathLike',
    'as_path',
    'ensure_directory',
    'ensure_suffix',
    'iter_files',
    'check_path_type',
    'get_user_config_path',
]




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
"""Core domain models and utilities for NEP workflows.

This package exposes lightweight, lazily-loaded entry points for the rest of
NepTrainKit to avoid heavy imports at startup and keep UI responsiveness high.

Notes
-----
- Public symbols are exported via ``__getattr__`` to defer imports.
- Modules in this package cover structures, messaging, dataset IO, and helpers.


"""
# Lightweight, lazy exports to avoid heavy imports at startup.
from __future__ import annotations

from typing import Any

__all__ = [
    'MessageManager',
    'CardManager', 'load_cards_from_directory',
]

from .card_manager import CardManager, load_cards_from_directory
from .message import MessageManager

# 可选：通过环境变量 NEP_NATIVE_STDIO 控制原生 stdout/stderr 重定向
try:
    import os as _os
    _mode = _os.environ.get("NEP_NATIVE_STDIO", "").strip()
    if _mode:
        from .cstdio_redirect import redirect_c_stdout_stderr as _redir
        if _mode.lower() == "inherit":
            _guard = None
        elif _mode.lower() == "silent":
            _guard = _redir(None)
            _guard.__enter__()
        else:
            _guard = _redir(_mode)
            _guard.__enter__()
        _NEP_NATIVE_STDIO_GUARD = _guard
except Exception:
    _NEP_NATIVE_STDIO_GUARD = None




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/14 17:17
# @Author  : 
# @email    : 1747193328@qq.com
"""Qt helpers for checking NepTrainKit updates and downloading assets."""

import json
import re
import sys
import traceback

from PySide6.QtCore import Signal, QObject
from loguru import logger
from qfluentwidgets import MessageBox

from NepTrainKit import is_nuitka_compiled, module_path
from NepTrainKit.core import MessageManager
from NepTrainKit.paths import as_path
from NepTrainKit.version import RELEASES_API_URL, __version__, UPDATE_FILE

from NepTrainKit.ui.threads import LoadingThread
from NepTrainKit.ui.updater import unzip


class UpdateWoker(QObject):
    """Worker that checks for new application releases and applies updates."""

    version = Signal(dict)
    download_success = Signal()

    def __init__(self, parent):
        """Initialise the worker with the owning ``parent`` widget."""
        self._parent = parent
        super().__init__(parent)
        self.func = self._check_update
        self.version.connect(self._check_update_call_back)
        self.download_success.connect(self._call_restart)
        self.update_thread = LoadingThread(self._parent, show_tip=False)
        self.down_thread = LoadingThread(self._parent, show_tip=True, title="Downloading")

    def download(self, url: str) -> None:
        """Download the update archive from ``url`` and emit completion."""
        import requests  # Imported lazily to keep startup fast.

        response = requests.get(url, stream=True)
        update_path = as_path(UPDATE_FILE)
        with update_path.open("wb") as target:
            for chunk in response.iter_content(1024):
                if chunk:
                    target.write(chunk)
        self.download_success.emit()

    def _call_restart(self) -> None:
        """Prompt the user to relaunch once the update archive is ready."""
        box = MessageBox(
            "Do you want to restart?",
            (
                "Update package downloaded successfully! Would you like to restart now?\n"
                "If you cancel, the update will be applied automatically the next time you open the software."
            ),
            self._parent,
        )
        box.yesButton.setText("Update")
        box.cancelButton.setText("Cancel")
        box.exec_()
        if box.result() == 0:
            return
        unzip()

    def _check_update(self) -> None:
        """Query GitHub releases and emit the response payload."""
        import requests

        MessageManager.send_info_message("Checking for updates, please wait...")
        try:
            headers = {"User-Agent": "Awesome-Octocat-App"}
            version_info = requests.get(RELEASES_API_URL, headers=headers).json()
            self.version.emit(version_info)
        except Exception:  # noqa: BLE001 - we log the traceback for user support.
            logger.error(traceback.format_exc())
            MessageManager.send_error_message("Network error!")

    def _check_update_call_back(self, version_info: dict) -> None:
        """Handle the GitHub API response and notify the user."""
        if "message" in version_info:
            MessageManager.send_warning_message(version_info["message"])
            return
        if version_info["tag_name"][1:] == __version__:
            MessageManager.send_success_message("You are already using the latest version!")
            return

        box = MessageBox(
            "New version detected:" + version_info["name"] + version_info["tag_name"],
            version_info["body"],
            self._parent,
        )
        box.yesButton.setText("Update")
        box.cancelButton.setText("Cancel")
        box.exec_()
        if box.result() == 0:
            return
        for assets in version_info["assets"]:
            if sys.platform in assets["name"] and "NepTrainKit" in assets["name"]:
                self.down_thread.start_work(self.download, assets["browser_download_url"])
                return
        MessageManager.send_warning_message("No update package available for your system. Please download it manually!")

    def check_update(self) -> None:
        """Start the background task if running from a packaged build."""
        if not is_nuitka_compiled:
            MessageManager.send_info_message("You can update via pip install NepTrainKit -U --pre")
            return
        self.update_thread.start_work(self._check_update)


class UpdateNEP89Woker(QObject):
    """Worker that keeps the bundled ``nep89`` potential file up to date."""

    version = Signal(int)
    download_success = Signal()

    def __init__(self, parent):
        """Initialise the worker with the owning ``parent`` widget."""
        self._parent = parent
        super().__init__(parent)
        self.func = self._check_update
        self.version.connect(self._check_update_call_back)
        self.update_thread = LoadingThread(self._parent, show_tip=False)
        self.down_thread = LoadingThread(self._parent, show_tip=True, title="Downloading")

    def download(self, latest_date: int) -> None:
        """Download the latest ``nep89`` model and refresh metadata."""
        import requests

        raw_url = (
            "https://raw.githubusercontent.com/brucefan1983/GPUMD/master/"
            f"potentials/nep/nep89_{latest_date}/nep89_{latest_date}.txt"
        )
        response = requests.get(raw_url, stream=True)
        nep89_path = module_path/ "Config/nep89.txt"
        with nep89_path.open("wb") as target:
            for chunk in response.iter_content(1024):
                if chunk:
                    target.write(chunk)

        MessageManager.send_success_message("Update large model completed!")
        nep_json_path = module_path/ "Config/nep.json"
        with nep_json_path.open("r", encoding="utf-8") as config_file:
            local_nep_info = json.load(config_file)
        local_nep_info["date"] = latest_date
        with nep_json_path.open("w", encoding="utf-8") as config_file:
            json.dump(local_nep_info, config_file)

    def _check_update(self) -> None:
        """Check the remote repository for a newer ``nep89`` dataset."""
        import requests

        MessageManager.send_info_message("Checking for updates, please wait...")
        api_url = "https://api.github.com/repos/brucefan1983/GPUMD/contents/potentials/nep"
        response = requests.get(api_url)
        if response.status_code != 200:
            MessageManager.send_warning_message(
                f"Unable to access the warehouse directory, status code: {response.status_code}"
            )
            return
        directories = [item["name"] for item in response.json() if item["type"] == "dir" and item["name"].startswith("nep89_")]

        date_pattern = re.compile(r"nep89_(\d{8})")
        latest_date: int | None = None
        for dir_name in directories:
            match = date_pattern.match(dir_name)
            if match:
                current_date = int(match.group(1))
                if latest_date is None or current_date > latest_date:
                    latest_date = current_date

        self.version.emit(latest_date)

    def _check_update_call_back(self, latest_date: int) -> None:
        """Prompt the user to download the updated ``nep89`` archive."""
        nep_json_path = module_path/ "Config/nep.json"
        with nep_json_path.open("r", encoding="utf-8") as config_file:
            local_nep_info = json.load(config_file)
        if local_nep_info["date"] >= latest_date:
            MessageManager.send_success_message("You are already using the latest version!")
            return
        box = MessageBox(
            "New version",
            f"A new version of the large model has been detected:{latest_date}",
            self._parent,
        )
        box.yesButton.setText("Update")
        box.cancelButton.setText("Cancel")
        box.exec_()
        if box.result() == 0:
            return
        self.down_thread.start_work(self.download, latest_date)

    def check_update(self) -> None:
        """Start checking for ``nep89`` updates in the background."""
        self.update_thread.start_work(self._check_update)

